{% load i18n %}
    {% if queryset.has_previous %}
        <input type="submit" name="first" value="&lt;&lt;" />
        <input type="submit" name="previous" value="&lt;" />
    {% else %}
        <input disabled type="submit" name="first" value="&lt;&lt;" />
        <input disabled type="submit" name="previous" value="&lt;" />
    {% endif %}
    ({% trans 'Page' %} {{ queryset.number }} {% trans 'of' %} {{ queryset.paginator.num_pages }}; {{ queryset.paginator.count }} {% trans 'items' %})
    {% if queryset.has_next %}
        <input type="submit" name="next" value="&gt;" />
        <input type="submit" name="last" value="&gt;&gt;" />
    {% else %}
        <input disabled type="submit" name="next" value="&gt;" />
        <input disabled type="submit" name="last" value="&gt;&gt;" />
    {% endif %}
