{% load admin_static %}{% load url from future %}<!DOCTYPE html>
<html lang="{{ LANGUAGE_CODE|default:"en-us" }}" {% if LANGUAGE_BIDI %}dir="rtl"{% endif %} xmlns="http://www.w3.org/1999/xhtml">
<head>
    <title>{{ bots_http_path }}</title>
    <link rel="stylesheet" type="text/css" href="{% block stylesheet %}{% static "admin/css/base.css" %}{% endblock %}" />
    {% block extrastyle %}{% endblock %}
    <!--[if lte IE 7]><link rel="stylesheet" type="text/css" href="{% block stylesheet_ie %}{{ STATIC_URL }}css/ie.css{% endblock %}" /><![endif]-->
    {% block extrahead %}{% endblock %}
    {% if LANGUAGE_BIDI %}<link rel="stylesheet" type="text/css" href="{% block stylesheet_rtl %}{{ STATIC_URL }}css/rtl.css{% endblock %}" />{% endif %}
    <script type="text/javascript">window.__admin_media_prefix__ = "{% filter escapejs %}{{ STATIC_URL }}{% endfilter %}";</script>
    <link rel="icon" type="image/png" href="/media/images/favicon.ico"/>
    <link rel="stylesheet" type="text/css" href="/media/css/bots.css" />
</head>
{% load i18n %}


<body class="{% if is_popup %}popup {% endif %}{% block bodyclass %}{% endblock %}">
{% if not is_popup %}
    {% include botslogo %}
    <div id="environment">
        <div style="color:{{ bots_environment_text_color }}">{{ bots_environment_text }}</div>
    </div>
    {% include "bots/menu.html" %}
{% endif %}

<div id="container">
    {% if messages %}
    <ul class="messages">
        {% for message in messages %}
        <li{% if message.tags %} class="{{ message.tags }}"{% endif %}>{{ message }}</li>
        {% endfor %}
    </ul>
    {% endif %}
    <div id="content" class="{% block coltype %}colM{% endblock %}">
        {% block pretitle %}{% endblock %}
        {% block content_title %}{% if title %}<h1>{{ title }}</h1>{% endif %}{% endblock %}
        {% block content %}
        {% block object-tools %}{% endblock %}
        {{ content }}
        {% endblock %}
        {% block sidebar %}{% endblock %}
        <br class="clear" />
    </div>
</div>
</body>
</html>
