{% extends "admin/base.html" %}
{% load i18n %}

{% block content %}
<br/>
<h1>Read a plugin</h1>
<p>&nbsp;Select a plugin from your file system:</p>
<br/>
<form action="/plugin" method="POST" enctype="multipart/form-data">{% csrf_token %}
    <table>
        {{ form }}
    </table>
    <br/>&nbsp;
    <input type="submit" name="cancel" value="{% trans 'Cancel' %}" />
    <input type="submit" name="submit" value="{% trans 'Read' %}" />
</form>
{% endblock %}
