{% load i18n %}
<div class="submit-row" {% if is_popup %}style="overflow: auto;"{% endif %}>
<!-- put buttons in reverse order that you want them to appear -->
                                  <input type="button" value="{% trans 'Cancel' %}"                    class="default" {% if is_popup %} onClick="window.close()" {% else %} onClick="history.back()" {% endif %}/>
{% if show_delete_link %}         <input type="button" value="{% trans 'Delete' %}"                    class="default" onclick="document.location.href='delete/'; return false;">{% endif %}
{% if show_save %}                <input type="submit" value="{% trans 'Save' %}"                      class="default" name="_save"       {{ onclick_attrib }}/>{% endif %}
{% if show_save_as_new %}         <input type="submit" value="{% trans 'Save as new' %}"               class="default" name="_saveasnew"  {{ onclick_attrib }}/>{% endif %}
{% if show_save_and_add_another %}<input type="submit" value="{% trans 'Save and add another' %}"      class="default" name="_addanother" {{ onclick_attrib }}/>{% endif %}
{% if show_save_and_continue %}   <input type="submit" value="{% trans 'Save and continue editing' %}" class="default" name="_continue"   {{ onclick_attrib }}/>{% endif %}
</div>
