div#logo {
	position:	absolute;
	left:		0%;
	width:		51em;
	top:		0em;
	height:		3.9em;
	}
div#logo img {
	position:	absolute;
	left:		0%;
	width:		100%;
	top:		0%;
	height:		100%;
	}
div#environment {
	position:	absolute;
	left:		52em;
	top:		1em;
	height:		2em;
	}
div#environment>div {
	font-size:	2em; 
	font-weight:bold; 
	}
div#menu {
	position:	absolute;
	top:		4.0em;
	height:		1.5em;
    z-index:	4;
	}
div#container {
	position:	absolute;
	top:		5.5em;
	}
div.help {
	position:	absolute;
	top:		5em;
	left:		5em;
	color:#000000;
	}
div#content {
	margin: 0;
	}
table#about {
	position:	absolute;
	left:		10%;
	top:		7em;
	}
/***********************************************************************/
/**MENU*****************************************************************/
div#menu a {
	color:      #000000;
	text-decoration:none;
	margin:	2px;
	}
div#menu ul {
	width:      8em;
	float:      left;
	margin:     0;
	padding:    0;
    }
div#menu hr {
    height: 1px;
	width: 100%;
	color: #bbb;
	background-color: #bbb;
	}
div#menu ul li {
	list-style: none;
	display:    none;
	font-size:  100%;
	}
div#menu ul:hover li, 
div#menu ul:hover li a {
	display:    block;
	width:      17em;
    border-left: 1px solid #ddd;
    color: #666;
    background: #efefef;
    border-bottom: 1px solid #efefef;
	padding:    0;
	margin:     0;
	}
div#menu ul:hover li a:hover, 
div#menu ul:hover a {
    color: #028;
    }
div#menu ul:hover li a:hover, 
div#menu ul:hover li a:hover {
    cursor:     pointer;
    }
/***********************************************************************/
/**ACTION**************************************************************/
div.dropdown {
	position:relative;
	float:left;
	}
div.dropdown  button {
    color: #666;
	}
div.dropdown  ul {
	list-style-type:none;
	margin:0;
	padding:0;
	position:absolute;
	visibility:hidden;
	left:99%;
	top:1px;
	background: #efefef;
	}
div.dropdown:hover  ul {
	visibility:visible;
	z-index:598;
	}
div.dropdown  ul  li {
	list-style: none;
    }
div.dropdown:hover button{
	background-color: #efefef;
	}
div.dropdown:hover >button,
div.dropdown  ul li:hover,
div.dropdown  ul li button:hover{
	color: #028;
	}
/***********************************************************************/
/**INFO**************************************************************/
div.info  table {
	position:absolute;
	visibility:hidden;
	background: #efefef;
    color: #666;
	white-space: nowrap;
	}
div.info:hover  table {
	visibility:visible;
	z-index:599;
	}
/***********************************************************************/
/**FILTER***************************************************************/
.change-list .filtered {
    background: none !important;
    }
.change-list .filtered .paginator,
.filtered .actions {
    margin-right:0 !important;
    }
div#changelist-filter h3, 
div#changelist-filter ul {
    display: none;
    }
div#changelist-filter:hover h3, 
div#changelist-filter:hover ul {
    display: block;
    }
/***********************************************************************/
/**other ***************************************************************/
.botsheader	{
	border-style:		none;
	margin:				0em;
	padding:		    0em;
	cursor:				pointer;
	white-space:		nowrap;
	color: 				#000000;
	background:			transparent;
	}
p.botsheader2	{
	border-style:		none;
	margin:				0em;
	padding:		    0em;
	white-space:		nowrap;
	color: 				#000000;
	font-weight:		normal;
	background:			#FFFFFF;
	}
a.botslink	{
	cursor:				pointer;
	white-space:		nowrap;
	text-decoration:	none;
    color:				#5b80b2;
    }
a.botslink:hover	{
    color:				#036;
    }
.nowrap	{
	white-space:		nowrap;
	}
table#botstable {
	border-collapse: collapse;
	border-width: 0;
	}
img.action {
	width: 1em;
	}
table#botstable td {
	line-height: 1em;
	}
.row1 {
	background: white;
	}
.row2 {
	background: #f8f8f8;
	}
/*nicer layout for sort boxes in admin for django1.4*/
table thead th .text span {
    padding: 15px 5px 2px 5px;
    display:block;
}
table thead th .text a {
    display: block;
    cursor: pointer;
    padding: 15px 5px 2px 5px;
}
